%{
let lineno = ref 0;;
%}

%token <int> INT
%token <string> VAR STRING
%token L_PRN R_PRN L_BRK R_BRK L_CRL R_CRL
%token SEMI COMMA MAIN FUNC
%token PLUS MINUS TIMES DIV MOD EQ NE LT LE GT GE AND OR NOT
%token ANDB ORB NOTB SHIFTR SHIFTL
%token ASSIGN MALLOC FREE
%token IF ELSE WHILE ASSERT INSPECTT RETURN BREAK CONTINUE
%token END POS OPENT SEEK READ
%token PRINT

%left AND OR
%nonassoc NOT
%left EQ NE LT LE GT GE
%left PLUS MINUS
%left ANDB ORB SHIFTR SHIFTL
%left TIMES DIV MOD
%nonassoc UMINUS NOTB

%start prog
%type <int * int> prog
/*(* (number of conditionals, cyclomatic complexity) *)*/

%%

prog:
  MAIN L_CRL stmts R_CRL
      {match $3 with (c, v) -> (c, v + 1)}  
  | FUNC VAR L_PRN R_PRN L_CRL stmts R_CRL prog
      {match ($6, $8) with ((c1, v1), (c2, v2)) -> ((c1 + c2), (v1 + v2 + 1))}
  | FUNC VAR L_PRN VAR R_PRN L_CRL stmts R_CRL prog
      {match ($7, $9) with ((c1, v1), (c2, v2)) -> ((c1 + c2), (v1 + v2 + 1))}
  | glo prog
      {$2}
  ;

glo:
  VAR ASSIGN INT SEMI                                     {(0, 0)}
  | VAR ASSIGN MALLOC L_PRN INT R_PRN SEMI                {(0, 0)}
  | VAR ASSIGN OPENT L_PRN STRING R_PRN SEMI              {(0, 0)}
  ;

stmts:
  /* empty */                 {(0, 0)}
  | stmt stmts                {match ($1, $2) with ((c1, v1), (c2, v2)) -> ((c1 + c2), (v1 + v2))}
  ;

exp:
  INT                         {(0, 0)}
  | VAR                       {(0, 0)}
  | L_PRN exp R_PRN           {$2}
  | exp PLUS exp              {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
      /* assume no one uses logical operations inside arithmetic expressions */
  | exp MINUS exp             {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | MINUS exp %prec UMINUS    {match $2 with (c, _) -> (c, 0)}
  | exp TIMES exp             {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp DIV exp               {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp MOD exp               {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp EQ exp                {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp NE exp                {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp LT exp                {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp LE exp                {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp GT exp                {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp GE exp                {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp AND exp               {match ($1, $3) with ((c1, v1), (c2, v2)) -> ((c1 + c2 + 1), (v1 + v2 + 1))}
  | exp OR exp                {match ($1, $3) with ((c1, v1), (c2, v2)) -> ((c1 + c2 + 1), (v1 + v2 + 1))}
  | NOT exp                   {$2}
  | exp ANDB exp              {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
      /* assume no one uses bitwise operations as logical */
  | exp ORB exp               {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | NOTB exp                  {match $2 with (c, _) -> (c, 0)}
  | exp SHIFTR exp            {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | exp SHIFTL exp            {match ($1, $3) with ((c1, _), (c2, _)) -> ((c1 + c2), 0)}
  | VAR L_BRK exp R_BRK       {match $3 with (c, _) -> (c, 0)}
  | END L_PRN VAR R_PRN       {(0, 0)}
  | POS L_PRN VAR R_PRN       {(0, 0)}
  ;

stmt:
  VAR ASSIGN exp SEMI                                                                   {$3}
  | VAR ASSIGN MALLOC L_PRN exp R_PRN SEMI                                              {$5}
  | FREE L_PRN VAR R_PRN SEMI                                                           {(0, 0)}
  | VAR L_BRK exp R_BRK ASSIGN exp SEMI
      {match ($3, $6) with ((c1, v1), (c2, v2)) -> ((c1 + c2), (v1 + v2))}
  | VAR ASSIGN VAR L_PRN R_PRN SEMI                                                     {(0, 0)}
  | VAR ASSIGN VAR L_PRN exp R_PRN SEMI                                                 {$5}
  | IF L_PRN exp R_PRN L_CRL stmts R_CRL ELSE L_CRL stmts R_CRL
      {match ($3, $6, $10) with ((c1, v1), (c2, v2), (c3, v3)) -> ((c1 + c2 + c3 + 1), (v1 + v2 + v3 + 1))}
  | IF L_PRN exp R_PRN L_CRL stmts R_CRL
      {match ($3, $6) with ((c1, v1), (c2, v2)) -> ((c1 + c2 + 1), (v1 + v2 + 1))}
  | WHILE L_PRN exp R_PRN L_CRL stmts R_CRL
      {match ($3, $6) with ((c1, v1), (c2, v2)) -> ((c1 + c2), (v1 + v2 + 1))}
  | ASSERT L_PRN exp R_PRN SEMI
      {match $3 with (c, v) -> (c + 1, v)}
  | INSPECTT L_PRN VAR COMMA INT R_PRN L_CRL stmts R_CRL
      {match $8 with (c, v) -> (c, v + 1)}
  | RETURN exp SEMI                                                                     {$2}
  | BREAK SEMI                                                                          {(0, 0)}
  | CONTINUE SEMI                                                                       {(0, 0)}
  | VAR ASSIGN OPENT L_PRN STRING R_PRN SEMI                                            {(0, 0)}
  | SEEK L_PRN VAR COMMA exp R_PRN SEMI                                                 {$5}
  | VAR ASSIGN READ L_PRN VAR R_PRN SEMI                                                {(0, 0)}
  | PRINT L_PRN exp R_PRN SEMI                                                          {$3}
  | SEMI                                                                                {(0, 0)}
  ;

%%
