type op_t =
  Plus
  | Minus
  | Times
  | Div
  | Mod
  | Eq
  | Ne
  | Lt
  | Le
  | Gt
  | Ge
  | And
  | Or
  | AndB
  | OrB
  | ShiftR
  | ShiftL

type exp_t =
  Int of int32
  | Var of string
  | BinOp of exp_t * op_t * exp_t
  | Not of exp_t
  | NotB of exp_t
  | ARead of string * exp_t
  | End of string * bool
  | Pos of string * bool
  | Valid of string

type dlm_t =
  DArr of string
  | DInt of int32

type stmt_t =
  VWrite of string * exp_t
  | Malloc of string * exp_t * bool
  | Free of string * bool
  | AWrite of string * exp_t * exp_t
  | Call0 of string * string
  | Call of string * string * exp_t
  | If of exp_t * stmts_t * stmts_t
  | While of exp_t * stmts_t
  | Return of exp_t
  | Break
  | Continue
  | OpenT of string * string * bool
  | Seek of string * string * exp_t * bool
  | Read of string * string * bool
  | Print of exp_t
  | Empty

and stmts_t = stmt_t list

type glo_t =
  FDefT of string * string * bool

type prog_t =
  Main of stmts_t
  | Fn0Def of string * stmts_t * prog_t
  | FnDef of string * string * stmts_t * prog_t
  | GloDef of glo_t * prog_t
