%{
open Types
let lineno = ref 0;;
%}

%token <int32> INT
%token <string> VAR STRING
%token L_PRN R_PRN L_BRK R_BRK L_CRL R_CRL
%token SEMI COMMA MAIN FUNC
%token PLUS MINUS TIMES DIV MOD EQ NE LT LE GT GE AND OR NOT
%token ANDB ORB NOTB SHIFTR SHIFTL
%token ASSIGN MALLOC FREE
%token IF ELSE WHILE ASSERT INSPECTT RETURN BREAK CONTINUE
%token END POS OPENT SEEK READ
%token PRINT

%left AND OR
%nonassoc NOT
%left EQ NE LT LE GT GE
%left PLUS MINUS
%left ANDB ORB SHIFTR SHIFTL
%left TIMES DIV MOD
%nonassoc UMINUS NOTB

%start prog
%type <Types.prog_t> prog

%%

prog:
  MAIN L_CRL stmts R_CRL
    {Main($3)}
  | FUNC VAR L_PRN R_PRN L_CRL stmts R_CRL prog
    {Fn0Def($2,$6,$8)}
  | FUNC VAR L_PRN VAR R_PRN L_CRL stmts R_CRL prog
    {FnDef($2,$4,$7,$9)}
  | glo prog
    {GloDef($1,$2)}
  ;

glo:
  VAR ASSIGN OPENT L_PRN STRING R_PRN SEMI
    {FDefT($1,$5,false)}
  ;

stmts:
  /* empty */             {[]}
  | stmt stmts       {$1::$2}
  ;

exp:
  INT                         {Int $1}
  | VAR                       {Var $1}
  | L_PRN exp R_PRN           {$2}
  | exp PLUS exp              {BinOp($1,Plus,$3)}
  | exp MINUS exp             {BinOp($1,Minus,$3)}
  | MINUS exp %prec UMINUS    {BinOp(Int Int32.zero,Minus,$2)}
  | exp TIMES exp             {BinOp($1,Times,$3)}
  | exp DIV exp               {BinOp($1,Div,$3)}
  | exp MOD exp               {BinOp($1,Mod,$3)}
  | exp EQ exp                {BinOp($1,Eq,$3)}
  | exp NE exp                {BinOp($1,Ne,$3)}
  | exp LT exp                {BinOp($1,Lt,$3)}
  | exp LE exp                {BinOp($1,Le,$3)}
  | exp GT exp                {BinOp($1,Gt,$3)}
  | exp GE exp                {BinOp($1,Ge,$3)}
  | exp AND exp               {BinOp($1,And,$3)}
  | exp OR exp                {BinOp($1,Or,$3)}
  | NOT exp                   {Not $2}
  | exp ANDB exp              {BinOp($1,AndB,$3)}
  | exp ORB exp               {BinOp($1,OrB,$3)}
  | NOTB exp                  {NotB $2}
  | exp SHIFTR exp            {BinOp($1,ShiftR,$3)}
  | exp SHIFTL exp            {BinOp($1,ShiftL,$3)}
  | VAR L_BRK exp R_BRK       {ARead($1,$3)}
  | END L_PRN VAR R_PRN       {End($3,false)}
  | POS L_PRN VAR R_PRN       {Pos($3,false)}
  ;

stmt:
  VAR ASSIGN exp SEMI
    {VWrite($1,$3)}
  | VAR ASSIGN MALLOC L_PRN exp R_PRN SEMI
    {Malloc($1,$5,false)}
  | FREE L_PRN VAR R_PRN SEMI
    {Free("",$3,false)}
  | VAR L_BRK exp R_BRK ASSIGN exp SEMI
    {AWrite($1,$3,$6)}
  | VAR ASSIGN VAR L_PRN R_PRN SEMI
    {Call0($1,$3)}
  | VAR ASSIGN VAR L_PRN exp R_PRN SEMI
    {Call($1,$3,$5)}
  | IF L_PRN exp R_PRN L_CRL stmts R_CRL ELSE L_CRL stmts R_CRL
    {If($3,$6,$10)}
  | IF L_PRN exp R_PRN L_CRL stmts R_CRL
    {If($3,$6,[])}
  | WHILE L_PRN exp R_PRN L_CRL stmts R_CRL
    {While($3,$6)}
  | ASSERT L_PRN exp R_PRN SEMI
    {Assert $3}
  | INSPECTT L_PRN VAR COMMA INT R_PRN L_CRL stmts R_CRL
    {InspectT((Int Int32.one),$3,(DInt $5),(DArr ""),$8)}
  | RETURN exp SEMI
    {Return($2)}
  | BREAK SEMI
    {Break}
  | CONTINUE SEMI
    {Continue}
  | VAR ASSIGN OPENT L_PRN STRING R_PRN SEMI
    {OpenT($1,$5,false)}
  | SEEK L_PRN VAR COMMA exp R_PRN SEMI
    {Seek("",$3,$5,false)}
  | VAR ASSIGN READ L_PRN VAR R_PRN SEMI
    {Read($1,$5,false)}
  | PRINT L_PRN exp R_PRN SEMI
    {Print $3}
  | SEMI
    {Empty}
  ;

%%
