let _ =
  if Array.length Sys.argv > 2 then
    (Printf.eprintf "Usage:\n\t%s\n\t%s source_file\n" Sys.argv.(0) Sys.argv.(0); exit 1)
  else
    let in_channel = if Array.length Sys.argv == 1 then stdin else open_in Sys.argv.(1) in
    let lexbuf = Lexing.from_channel in_channel in
    try
      while true do
        let (conds, cyclo) = Measure.prog Lexer.token lexbuf in (* parsing must come first *)
        let stmts = !Lexer.stmtno in (* meaningful statements *)
        let comps = !Lexer.compno in (* unconditional computations *)
        Printf.printf "Cyclomatic complexity: %d\n" cyclo;
        Printf.printf "Conditional clauses: %d\n" conds;
        Printf.printf "Statements overall: %d\n" stmts;
        Printf.printf "Statements manipulating data: %d\n" comps;
        flush stderr; flush stdout
      done
    with
      Lexer.Eof -> close_in in_channel; exit 0
    | Parsing.Parse_error ->
        let line = !Lexer.lineno in
        let tok = Lexing.lexeme lexbuf in
        Printf.eprintf "Syntax error: \"%s\" (line %d)\n" tok line;
        exit 1
;;
