{
open Measure
exception Eof
module StringSet = Set.Make(String);;
let lineno = ref 1;; (* lines of source code *)
let stmtno = ref 0;; (* meaningful statements, subset of lineno *)
let compno = ref 0;; (* unconditional computations, subset of stmtno *)
}

let numeric = ['0' - '9']
let letter = ['a' - 'z' 'A' - 'Z']
let strchar = [^ '\n' '"']
let chchar = [^ '\t' '\n' '\'']
let comment = "//" [^ '\n']* '\n'
let variable = (letter | '_') (letter | numeric | '_')*

rule token = parse
  [' ' '\t']                    {token lexbuf}
  | ['\n']                      {lineno := !lineno + 1; token lexbuf}
  | comment                     {lineno := !lineno + 1; token lexbuf}
  | numeric+ as num             {INT(int_of_string num)}
  | "'\\n'"                     {INT(int_of_char '\n')}
  | "'\\t'"                     {INT(int_of_char '\t')}
  | "'\\\\'"                    {INT(int_of_char '\\')}
  | "'" (chchar as ch) "'"      {INT(int_of_char ch)}
  | '"' (strchar+ as str) '"'   {STRING(str)}
  | '('                         {L_PRN}
  | ')'                         {R_PRN}
  | '['                         {L_BRK}
  | ']'                         {R_BRK}
  | '{'                         {L_CRL}
  | '}'                         {R_CRL}
  | ';'                         {SEMI}
  | ','                         {COMMA}
  | "main"                      {stmtno := !stmtno + 1; compno := !compno + 1; MAIN}
  | "func"                      {stmtno := !stmtno + 1; compno := !compno + 1; FUNC}
  | '+'                         {PLUS}
  | '-'                         {MINUS}
  | '*'                         {TIMES}
  | '/'                         {DIV}
  | '%'                         {MOD}
  | "=="                        {EQ}
  | "!="                        {NE}
  | '<'                         {LT}
  | "<="                        {LE}
  | '>'                         {GT}
  | ">="                        {GE}
  | "&&"                        {AND}
  | "||"                        {OR}
  | '!'                         {NOT}
  | '&'                         {ANDB}
  | '|'                         {ORB}
  | '~'                         {NOTB}
  | ">>"                        {SHIFTR}
  | "<<"                        {SHIFTL}
  | '='                         {stmtno := !stmtno + 1; compno := !compno + 1; ASSIGN}
  | "malloc"                    {MALLOC}
  | "malloc_ec"                 {MALLOC_EC}
  | "free"                      {stmtno := !stmtno + 1; compno := !compno + 1; FREE}
  | "free_ec"                   {FREE_EC}
  | "if"                        {stmtno := !stmtno + 1; IF}
  | "else"                      {ELSE}
  | "while"                     {stmtno := !stmtno + 1; compno := !compno + 1; WHILE}
  | "assert"                    {stmtno := !stmtno + 1; ASSERT}
  | "inspectt"                  {stmtno := !stmtno + 1; compno := !compno + 1; INSPECTT}
  | "lookatt"                   {stmtno := !stmtno + 1; compno := !compno + 1; LOOKATT}
  | "inspectb"                  {stmtno := !stmtno + 1; compno := !compno + 1; INSPECTB}
  | "lookatb"                   {stmtno := !stmtno + 1; compno := !compno + 1; LOOKATB}
  | "try"                       {stmtno := !stmtno + 1; compno := !compno + 1; TRY}
  | "catch"                     {CATCH}
  | "return"                    {stmtno := !stmtno + 1; compno := !compno + 1; RETURN}
  | "break"                     {stmtno := !stmtno + 1; compno := !compno + 1; BREAK}
  | "continue"                  {stmtno := !stmtno + 1; compno := !compno + 1; CONTINUE}
  | "end"                       {END}
  | "end_ec"                    {END_EC}
  | "pos"                       {POS}
  | "pos_ec"                    {POS_EC}
  | "size"                      {SIZE}
  | "size_ec"                   {SIZE_EC}
  | "opent"                     {OPENT}
  | "opent_ec"                  {OPENT_EC}
  | "openb"                     {OPENB}
  | "openb_ec"                  {OPENB_EC}
  | "seek"                      {stmtno := !stmtno + 1; compno := !compno + 1; SEEK}
  | "seek_ec"                   {SEEK_EC}
  | "read"                      {READ}
  | "read_ec"                   {READ_EC}
  | "valid"                     {VALID}
  | "debug"                     {DEBUG}
  | "print"                     {stmtno := !stmtno + 1; compno := !compno + 1; PRINT}
  | "exit"                      {stmtno := !stmtno + 1; compno := !compno + 1; EXIT}
  | variable as id              {VAR(id)}
  | eof                         {raise Eof}

