{
open Parser
exception Eof
exception Unexpected
let lineno = ref 1;;
}

let numeric = ['0' - '9']
let letter = ['a' - 'z' 'A' - 'Z']
let strchar = [^ '\n' '"']
let chchar = [^ '\t' '\n' '\'']
let comment = "//" [^ '\n']* '\n'
let variable = (letter | '_') (letter | numeric | '_')*

rule token = parse
  [' ' '\t']                    {token lexbuf}
  | ['\n']                      {lineno := !lineno + 1; token lexbuf}
  | comment                     {lineno := !lineno + 1; token lexbuf}
  | numeric+ as num             {INT(Int32.of_string num)}
  | "'\\n'"                     {INT(Int32.of_int (int_of_char '\n'))}
  | "'\\t'"                     {INT(Int32.of_int (int_of_char '\t'))}
  | "'\\\\'"                    {INT(Int32.of_int (int_of_char '\\'))}
  | "'" (chchar as ch) "'"      {INT(Int32.of_int (int_of_char ch))}
  | '"' (strchar+ as str) '"'   {STRING(str)}
  | '('                         {L_PRN}
  | ')'                         {R_PRN}
  | '['                         {L_BRK}
  | ']'                         {R_BRK}
  | '{'                         {L_CRL}
  | '}'                         {R_CRL}
  | ';'                         {SEMI}
  | ','                         {COMMA}
  | "main"                      {MAIN}
  | "func"                      {FUNC}
  | '+'                         {PLUS}
  | '-'                         {MINUS}
  | '*'                         {TIMES}
  | '/'                         {DIV}
  | '%'                         {MOD}
  | "=="                        {EQ}
  | "!="                        {NE}
  | '<'                         {LT}
  | "<="                        {LE}
  | '>'                         {GT}
  | ">="                        {GE}
  | "&&"                        {AND}
  | "||"                        {OR}
  | '!'                         {NOT}
  | '&'                         {ANDB}
  | '|'                         {ORB}
  | '~'                         {NOTB}
  | ">>"                        {SHIFTR}
  | "<<"                        {SHIFTL}
  | '='                         {ASSIGN}
  | "malloc"                    {MALLOC}
  | "malloc_ec"                 {MALLOC_EC}
  | "free"                      {FREE}
  | "free_ec"                   {FREE_EC}
  | "if"                        {IF}
  | "else"                      {ELSE}
  | "while"                     {WHILE}
  | "assert"                    {ASSERT}
  | "inspectt"                  {INSPECTT}
  | "lookatt"                   {LOOKATT}
  | "inspectb"                  {INSPECTB}
  | "lookatb"                   {LOOKATB}
  | "try"                       {TRY}
  | "catch"                     {CATCH}
  | "return"                    {RETURN}
  | "break"                     {BREAK}
  | "continue"                  {CONTINUE}
  | "end"                       {END}
  | "end_ec"                    {END_EC}
  | "pos"                       {POS}
  | "pos_ec"                    {POS_EC}
  | "size"                      {SIZE}
  | "size_ec"                   {SIZE_EC}
  | "opent"                     {OPENT}
  | "opent_ec"                  {OPENT_EC}
  | "openb"                     {OPENB}
  | "openb_ec"                  {OPENB_EC}
  | "seek"                      {SEEK}
  | "seek_ec"                   {SEEK_EC}
  | "read"                      {READ}
  | "read_ec"                   {READ_EC}
  | "valid"                     {VALID}
  | "debug"                     {DEBUG}
  | "print"                     {PRINT}
  | "exit"                      {EXIT}
  | variable as id              {VAR(id)}
  | eof                         {raise Eof}
  | _                           {raise Unexpected}

